<?php

namespace apexl\Vault\interfaces;

use stdClass;

interface operators
{
    public function execute(): operators;

    public function executeRaw(string $sql): array;

    public function fetch(): stdClass|false;

    public function fetchColumn(int $ix = 0): mixed;

    public function fetchAll(): array|false;

    public function fetchAllAssoc(): array|false;

    public function fetchAssoc(): array|false;

    public function setLoggedInUserId(int $userId);

    public function isCount(): bool;

    /**
     * @return string[]
     */
    public function getQueryFields(): array;
}
